import os
import shutil
import unittest

from rss import api

test_header = """C01 CLIENT    : WOODSIDE ENERGY LTD                                             
C02 SURVEY    : ENFIELD 2010 4D MONITOR 5, BLOCK WA-28-L, NW AUSTRALIAN         
C03 DATASET   : FINAL FAR ANGLE (30-41) STACK OF MONITOR 5 IN TIME DOMAIN       
C04 INLINES   : 1000-1253 (INCR 1), CROSSLINES: 1430-2521 (INCR 1)              
C05 SRATE     : 4MS, TMAX : 4000MS                                              
C06 GRID      : 12.5 X 25 M                                                     
C07 TAPELABEL : GU5133                                                          
C08 RECORDING PARAMETERS:                                                       
C09 SHOT INTERVAL : 18.75M  RCV INTERVAL    : 12.50M                            
C10 CABLE LENGTH  : 2800M   RCVS PER CABLE  : 224 GUNS/CABLES : 2/8             
C11 GUN SEPARATION: 50.0M   CABLE SEPARATION: 100.0M                            
C12 MIN OFFSET: 580 M       MAX OFFSET: 3380 M                                  
C13 NOTE: THIS IS A 4D PUSH-REVERSE SURVEY WITH SOURCE BOAT BEHIND TAILBUOYS    
C14 DATA ORDERED BY FIELDRECORD AND CHANNEL NUMBER-CHANNEL 1 IS FAR OFFSET CHAN 
C15 SPHEROID OF REF: GDA94  UTM ZONE:50S COORDINATE UNITS: METERS               
C16 PROCESSED DATUM: MSL SAMPLE RATE: 2MS  MAXTIME: 4096MS                      
C17 PROCESSING SEQUENCE:(WESTERN)                                               
C18 CONVERT SEGD TO WESTERN FORMAT, 2MS SAMPLE RATE, 4096MS TRACE LENGTH        
C19 FLAG FIELD EDITS, SEISMIC-NAVIGATION MERGE,UPDATE TRACE HEADERS,SEGY OUTPUT 
C20 PROCESSING SEQUENCE:(CGGVERITAS)                                            
C21 REFORMAT FROM SEGY TO CGGVERITAS INTERNAL FORMAT                            
C22 WOODSIDE SUPPLIED ANTI ALIAS FILTER AND RESAMPLE TO 4 MS                    
C23 ONE SAMPLE TIME SHIFT (4MS SHIFT COMPARING WITH WESTERN FORMAT)             
C24 ZERO PHASE LOW-CUT FILTER (3HZ/18DB) AND ZERO PHASE DESIGNATURE             
C25 GUN/CABLE STATICS CORRECTION (5M/7M, VELOCITY: 1500 M/S) & TIDAL STATICS    
C26 SCALE, TIME & PHASE SHIFT CORRECTION (SCALAR:1.4,SHIFT:6.799MS/-12.326DEG)  
C27 CUT OFFSET INTO RANGE 587-3213 (BIN CENTER 625-3175)                        
C28 FIRST PASS SWELL NOISE ATTENUATION & LINEAR NOISE REMOVAL                   
C29 SECOND PASS SWELL NOISE ATTENUATION & EXTRA DENOISE FOR 16 VERY NOISY LINES 
C30 3D SRME, RANDOMISED GATHER DENOISE, RADON DEMULTIPLE & INVERSE Q            
C31 SHOT & CHANNEL SCALING, RMC & WATER COLUMN STATICS CORRECTION               
C32 GLOBAL MATCH, MERGE WITH BASE & MISSING TRACE INTERPOLATION                 
C33 4D DSRNRMS BINNING (6.25M X 25M GRID),DEBUBBLE,PSDM & ADJACENT BINS SUM     
C34 RESIDUAL MOVEOUT CORRECTION & STRETCH TO TIME WITH SMOOTH VELOCITY          
C35 RADON DEMULTIPLE,OFFSET DEPENDENT SPECTRAL BALANCE,TRIM STATICS,DENOISE     
C36 WOODSIDE SPECTRAL BALANCE, ANGLE MUTE (30-41), STACK AND WAVELET DENOISE    
C37 GLOBAL MATCH WITH BASE,FULL FOLD POLYGON CUT & OUTPUT IN SEGY FORMAT        
C38 CDP       :BYTE 21-24    TRACE SEQ :BYTE 25-28    ARCHV NO:BYTE 173-174     
C39 CDPX(M)   :BYTE 181-184  CDPY(M)   :BYTE 185-188 (MULTIPLIED BY 100)        
C40 INLINE    :BYTE 189-192  CROSSLINE :BYTE 193-196                            
"""

binary_meta = {
    "sample_rate_ms": 4.0,
    "ns": 1001,
    "float_format": 1,
    "units": "meters",
    "num_traces": 1,
    "size_of_trace": 4244,
}


class TestAPI(unittest.TestCase):
    def setUp(self):
        self.segy_file = "data/enfield_test_data.sgy"
        self.ieee_data = "data/psdn_test_data.segy"

    def test_ebcdic(self):
        from rss.api import parse_ebcdic

        # just print for QC
        result = parse_ebcdic(self.segy_file)
        self.assertEqual(test_header, result)

    def test_parse_binary_header(self):
        from rss.api import parse_binary_header

        binary_header = parse_binary_header(self.segy_file)
        for key, val in binary_header.items():
            self.assertEqual(binary_meta[key], val)

    def test_read_trace_data_ieee(self):
        import numpy as np
        from rss.api import read_trace_data

        binary_meta = {
            "sample_rate_ms": 1.0,
            "ns": 1501,
            "float_format": 5,
            "units": "meters",
            "num_traces": 120,
            "size_of_trace": 240 + 4 * 1501,
        }

        read_size = binary_meta["size_of_trace"] * binary_meta["num_traces"]
        with open(self.ieee_data, "rb") as fp:
            fp.seek(3600)
            traces = np.frombuffer(fp.read(read_size), ">f")

        traces.shape = (120, -1)
        traces = traces[:, 60:]

        read_trace_data(self.ieee_data, binary_meta)

        output_file = os.path.join(
            os.path.splitext(os.path.basename(self.ieee_data))[0],
            "inlines",
            "983",
            "traces.bin",
        )

        with open(output_file, "rb") as fp:
            test_traces = np.frombuffer(fp.read(1501 * 4 * 120), "<f")
        test_traces.shape = (120, -1)

        try:
            np.testing.assert_array_equal(traces, test_traces)
            shutil.rmtree(
                os.path.splitext(os.path.basename(self.ieee_data))[0]
            )
        except:
            shutil.rmtree(
                os.path.splitext(os.path.basename(self.ieee_data))[0]
            )
            raise

    def test_read_trace_data_ibm(self):
        from ibm2ieee import ibm2float32
        import numpy as np

        from rss.api import read_trace_data

        with open(self.segy_file, "rb") as fp:
            fp.seek(3600 + 240)
            trace = ibm2float32(np.frombuffer(fp.read(1001 * 4), dtype=">u4"))

        read_trace_data(self.segy_file, binary_meta, scalco=-100)

        output_file = os.path.join(
            os.path.splitext(os.path.basename(self.segy_file))[0],
            "inlines",
            "1253",
            "traces.bin",
        )

        with open(output_file, "rb") as fp:
            test_trace = np.frombuffer(fp.read(1001 * 4), "<f")

        np.testing.assert_array_equal(trace, test_trace)

        try:
            np.testing.assert_array_equal(trace, test_trace)
            shutil.rmtree(
                os.path.splitext(os.path.basename(self.segy_file))[0]
            )
        except:
            shutil.rmtree(
                os.path.splitext(os.path.basename(self.segy_file))[0]
            )
            raise

    def test_compressed_zarr(self):
        from ibm2ieee import ibm2float32
        import numpy as np
        from rss.api import compressed_zarr, read_trace_data
        from rss.client import rssFromFile

        binary_meta = {
            "sample_rate_ms": 1.0,
            "ns": 1501,
            "float_format": 5,
            "units": "meters",
            "num_traces": 120,
            "size_of_trace": 240 + 4 * 1501,
        }

        read_trace_data(self.ieee_data, binary_meta, sort_order="inline")
        read_trace_data(self.ieee_data, binary_meta, sort_order="crossline")

        compressed_zarr(self.ieee_data, sort_order="inline")
        compressed_zarr(self.ieee_data, sort_order="crossline")

        # test the client too
        rss = rssFromFile("psdn_test_data")

        traces, _ = rss.line(983, sort_order="inline")

        np.testing.assert_array_equal(traces.shape, (1501, 120))
        # crossline will be the same because we only have one trace

        with open(self.ieee_data, "rb") as fp:
            fp.seek(3600)
            read_size = (
                binary_meta["size_of_trace"] * binary_meta["num_traces"]
            )
            test_traces = np.frombuffer(fp.read(read_size), dtype=">f4")

        test_traces.shape = (120, -1)
        test_traces = test_traces[:, 60:].T

        # test ranges:
        np.testing.assert_allclose(traces.max(), test_traces.max(), rtol=1e-4)

        np.testing.assert_allclose(traces.min(), test_traces.min(), rtol=1e-4)

        # the largest relative error is where the amplitude is smallest
        dynamic_range = test_traces.max() - test_traces.min()
        delta = np.abs((traces - test_traces) / dynamic_range)
        self.assertTrue(delta.max() < 1e-4)

        shutil.rmtree(os.path.splitext(os.path.basename(self.ieee_data))[0])



from collections import defaultdict
import ebcdic
from itertools import chain
from glob import glob
from ibm2ieee import ibm2float32
import json
from numcodecs import LZ4
import numpy as np
import os
import struct
import tqdm
import zarr

# SEGY definitions
headers_offset = 3600
trace_header_size = 240
size_of_float = 4

segy_units = {0: "unknown", 1: "meters", 2: "feet"}

segy_format = {
    1: "4-byte IBM floating-point",
    2: "4-byte, twos complement integer",
    3: "2-byte, twos complement integer",
    4: "4-byte fixed-point with gain",
    5: "4-byte IEEE floating-point",
    6: "unknown",
    7: "unknown",
    8: "1-byte twos complement integer",
}

byte_locations = {
    "inline": (189, 4, ">i"),
    "crossline": (193, 4, ">i"),
    "cdpx": (181, 4, ">i"),
    "cdpy": (185, 4, ">i"),
    "scalco": (71, 2, ">h"),
}

# default compression
compressor = LZ4()


def parse_ebcdic(segy_file):
    with open(segy_file, "rb") as fp:
        ebcdic_bytes = fp.read(3200).decode("cp1140")
    ebcdic_header = ""
    for i in range(40):
        ebcdic_header += ebcdic_bytes[i * 80 : (i + 1) * 80] + "\n"
    return ebcdic_header


def parse_binary_header(segy_file):
    with open(segy_file, "rb") as fp:
        fp.seek(3200)
        binary_header = fp.read(400)

    sample_rate_ms = struct.unpack(">H", binary_header[16:18])[0] / 1000.0
    ns = struct.unpack(">H", binary_header[20:22])[0]
    float_format = struct.unpack(">H", binary_header[24:26])[0]
    spatial_units = struct.unpack(">H", binary_header[54:56])[0]

    if float_format not in (1, 5):
        fmt = segy_format[float_format]
        raise RuntimeError(f"binary format {fmt} not supported.")

    size_of_trace = ns * 4 + 240

    file_size = os.path.getsize(segy_file)

    if (file_size - 3600) % size_of_trace:
        raise RuntimeError("Variable trace length not supported.")

    return {
        "sample_rate_ms": sample_rate_ms,
        "ns": ns,
        "float_format": float_format,
        "units": segy_units[spatial_units],
        "size_of_trace": size_of_trace,
        "num_traces": (file_size - 3600) // size_of_trace,
    }


def to_coord(x, scal):
    if scal > 0:
        return x * scal
    else:
        return x / abs(scal)


def parse_header(trace_as_bytes, scalco):
    hdr = trace_as_bytes[:trace_header_size]

    hdr = {
        key: struct.unpack(val[2], hdr[val[0] - 1 : val[0] - 1 + val[1]])[0]
        for key, val in byte_locations.items()
    }

    if scalco is None:
        scalco = hdr["scalco"]
    # convert to scale
    hdr["cdpx"] = to_coord(hdr["cdpx"], scalco)
    hdr["cdpy"] = to_coord(hdr["cdpy"], scalco)

    return hdr


def parse_trace(trace_as_bytes, binary_format):
    trace_data = trace_as_bytes[trace_header_size:]

    if binary_format == 1:
        trace_data = ibm2float32(np.frombuffer(trace_data, dtype=">u4"))
    elif binary_format == 5:
        trace_data = np.frombuffer(trace_data, dtype=">f")
        trace_data = trace_data.byteswap()
    else:
        fmt = segy_format[float_format]
        raise RuntimeError(f"binary format {fmt} not supported.")
    return trace_data


def read_trace_data(
    segy_file,
    binary_header,
    byte_locations=byte_locations,
    scalco=None,
    sort_order="inline",
):
    sort_order = sort_order.lower()
    if sort_order not in ("inline", "crossline"):
        raise RuntimeError(
            f"{sort_order} not supported, sort order should be on of inline or crossline."
        )

    if sort_order == "inline":
        orthogonal_line = "crossline"
    else:
        orthogonal_line = "inline"

    filename = os.path.splitext(os.path.basename(segy_file))[0]

    folder = f"{filename}"
    if not os.path.exists(folder):
        os.makedirs(folder)

    with open(os.path.join(filename, "binary_header.json"), "w") as fp:
        fp.write(json.dumps(binary_header))

    trace_size = binary_header["size_of_trace"]

    line_coords = defaultdict(list)

    inlines = np.zeros(binary_header["num_traces"], dtype=int)
    crosslines = np.zeros(binary_header["num_traces"], dtype=int)

    cdpx = np.zeros(binary_header["num_traces"], dtype=int)
    cdpy = np.zeros(binary_header["num_traces"], dtype=int)

    with open(segy_file, "rb") as fp:
        fp.seek(3600)
        for trac in tqdm.tqdm(range(0, binary_header["num_traces"])):
            raw_bytes = fp.read(binary_header["size_of_trace"])

            hdr = parse_header(raw_bytes, scalco)
            trace = parse_trace(raw_bytes, binary_header["float_format"])

            # Dumbest possible impl
            line_number = hdr[sort_order]
            folder = os.path.join(filename, f"{sort_order}s", f"{line_number}")
            if not os.path.exists(folder):
                os.makedirs(folder)
            with open(os.path.join(folder, "traces.bin"), "ba") as gp:
                trace.tofile(gp)
            line_coords[hdr[sort_order]].append(hdr[orthogonal_line])

            # save all the inlines
            inlines[trac] = hdr["inline"]
            crosslines[trac] = hdr["crossline"]

            cdpx[trac] = hdr["cdpx"]
            cdpy[trac] = hdr["cdpy"]

    np.save(os.path.join(filename, "inlines.npy"), inlines)
    np.save(os.path.join(filename, "crosslines.npy"), crosslines)

    np.save(os.path.join(filename, "cdpx.npy"), cdpx)
    np.save(os.path.join(filename, "cdpy.npy"), cdpy)

    inlines = np.unique(inlines)
    crosslines = np.unique(crosslines)

    min_inline = inlines[inlines > 0].min()
    max_inline = inlines[inlines > 0].max()

    min_crossline = crosslines[crosslines > 0].min()
    max_crossline = crosslines[crosslines > 0].max()

    num_inlines = len(inlines)
    num_crosslines = len(crosslines)

    if sort_order == "inline":
        min_orth_line = min_crossline
        num_orth_lines = num_crosslines
    else:
        min_orth_line = min_inline
        num_orth_lines = num_inlines

    # seems backwards, but the "crosslines" are in the "inline" image
    # i.e. the inline number is constant and vice versa:
    for key, val in line_coords.items():
        indx = np.array(val) - min_orth_line
        valid_crosslines = np.zeros(num_orth_lines, dtype=bool)
        valid_crosslines[indx] = 1

        output_file = os.path.join(
            filename, f"{sort_order}s", f"{key}", "index.bin"
        )
        with open(output_file, "ba") as gp:
            valid_crosslines.tofile(gp)


def compressed_zarr(segy_file, sort_order="inline"):
    sort_order = sort_order.lower()
    if sort_order not in ("inline", "crossline"):
        raise RuntimeError(
            f"{sort_order} not supported, sort order should be on of inline or crossline."
        )

    if sort_order == "inline":
        orthogonal_line = "crossline"
    else:
        orthogonal_line = "inline"

    filename = os.path.splitext(os.path.basename(segy_file))[0]

    with open(os.path.join(filename, "binary_header.json"), "r") as fp:
        binary_header = json.loads(fp.read())

    cdpx = np.load(os.path.join(filename, "cdpx.npy"))
    cdpy = np.load(os.path.join(filename, "cdpy.npy"))

    inlines = np.load(os.path.join(filename, "inlines.npy"))
    crosslines = np.load(os.path.join(filename, "crosslines.npy"))

    if sort_order == "inline":
        min_orth_line = crosslines.min()
        max_orth_line = crosslines.max()

        min_line = inlines.min()
        max_line = inlines.max()
    else:
        min_orth_line = inlines.min()
        max_orth_line = inlines.max()

        min_line = crosslines.min()
        max_line = crosslines.max()

    # always read whole traces:

    chunks = [
        int(binary_header["ns"]),
        int(max_orth_line - min_orth_line + 1),
        1,
    ]

    # folder = f'{filename}/traces.zarr'
    store = zarr.DirectoryStore(f"{filename}")
    root = zarr.group(store)

    bounds = root.create_dataset(
        "bounds",
        data=[
            int(inlines.min()),
            int(crosslines.min()),
            int(inlines.max()),
            int(crosslines.max()),
        ],
        dtype=int,
        overwrite=True,
    )

    coords_root = root.create_group("coords", overwrite=True)

    inline_coord = coords_root.create_dataset(
        "inlines", data=inlines, dtype=int, overwrite=True
    )
    crossline_coord = coords_root.create_dataset(
        "crosslines", data=crosslines, dtype=int, overwrite=True
    )
    cdpx_coord = coords_root.create_dataset(
        "cdpx", data=cdpx, dtype=float, overwrite=True
    )
    cdpy_coord = coords_root.create_dataset(
        "cdpy", data=cdpy, dtype=float, overwrite=True
    )

    line_root = root.create_group(sort_order, overwrite=True)

    seismic = line_root.zeros(
        "seismic",
        shape=(
            binary_header["ns"],
            max_orth_line - min_orth_line + 1,
            max_line - min_line + 1,
        ),
        chunks=chunks,
        compressor=compressor,
        dtype=np.uint16,
        overwrite=True,
    )

    scalers = line_root.zeros(
        "scalers", shape=(max_line - min_line + 1, 2), dtype=int
    )

    folder = os.path.join(filename, f"{sort_order}s", "*")
    for line in tqdm.tqdm(glob(folder)):
        line_number = int(os.path.basename(line))

        # FIXME rerun little endian
        traces = np.fromfile(
            os.path.join(
                filename, f"{sort_order}s", f"{line_number}", "traces.bin"
            ),
            dtype="<f4",
        )
        indx = np.fromfile(
            os.path.join(
                filename, f"{sort_order}s", f"{line_number}", "index.bin"
            ),
            dtype=bool,
        )

        min_val = traces.min()
        traces -= min_val

        max_val = traces.max()
        if max_val != 0:
            # could the entire line be zero?
            traces = (65535 - 1) * traces / max_val

        # zero isn't an invalid number
        traces += 1

        traces.shape = (-1, binary_header["ns"])

        _traces = np.zeros(
            (binary_header["ns"], max_orth_line - min_orth_line + 1),
            dtype=np.uint16,
        )

        _traces[:, np.where(indx)[0]] = traces.T

        seismic[..., line_number - min_line] = _traces

        scalers[line_number - min_line, :] = [min_val, max_val]



from setuptools import setup, find_packages
from os import path
from io import open

here = path.abspath(path.dirname(__file__))


setup(
    name="real-simple-seismic",
    version="0.0.2",  # Required
    description="real simple seismic",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3.6",
    ],
    packages=find_packages(exclude=["contrib",
                                    "docs",
                                    "tests"]),
    python_requires=">=3.6",

    # This is for getting through appveyor, install tensorflow_gpu if you can.
    install_requires=["numpy", 
                      "ebcdic", 
                      "ibm2ieee", 
                      "tqdm",
                      'numcodecs', 
                      's3fs',
                      "scipy",
                      "zarr"],  
    # List additional groups of dependencies here (e.g. development
    # dependencies). Users will be able to install these using the "extras"
    # syntax, for example:
    #
    #   $ pip install sampleproject[dev]
    #
    # Similar to `install_requires` above, these must be valid existing
    # projects.
    extras_require={},
    # If there are data files included in your packages that need to be
    # installed, specify them here.
    #
    # If using Python 2.6 or earlier, then these have to be included in
    # MANIFEST.in as well.
    package_data={},
    # To provide executable scripts, use entry points in preference to the
    # "scripts" keyword. Entry points provide cross-platform support and allow
    # `pip` to create the appropriate form of executable for the target
    # platform.
    entry_points={},
    # List additional URLs that are relevant to your project as a dict.
    project_urls={},
)



from IPython.display import clear_output
import numpy as np
import os
import s3fs
from scipy.spatial import KDTree
import zarr


def load_trace(seismic, scalers, bounds, inline, crossline):
    """
    Loads a trace from the input seismic array.

    Parameters
    ----------
    seismic : array like object containing sort order.
    scalers : array like object containing dynamic range of the line.
    bounds : dict containing min/max values for the inline/crossline coords.
    inline : int, the inline number to access.
    crossline : int, the crossline number to access.

    Returns
    -------
    trace : 1-D float array containing the trace data.
    is_live : boolean, is this a live trace or one added by padding.
    """

    min_inline, min_crossline, max_inline, max_crossline = bounds

    if inline < min_inline or inline > max_inline:
        raise RuntimeError(
            f"{inline} out of bounds [{min_inline}, {max_inline}]."
        )

    if crossline < min_crossline or crossline > max_crossline:
        raise RuntimeError(
            f"{crossline} out of bounds [{min_crossline}, {max_crossline}]."
        )

    trace = seismic[:, crossline - min_crossline, inline - min_inline]
    mask = trace < 1
    min_val, max_val = scalers[inline - min_inline, :]
    trace -= 1
    trace = trace.astype(float) * max_val / (65535 - 1)
    trace += min_val

    is_live = np.all(~mask)

    return trace, mask


def load_line(
    seismic, scalers, bounds, line_number, mask_val=np.nan, sort_order="inline"
):
    """
    Loads a line from the input seismic array and resizes it to a standardized size, with
    constant padding.

    Parameters
    ----------
    seismic : array like object containing sort order.
    scalers : array like object containing dynamic range of the line.
    bounds : dict containing min/max values for the inline/crossline coords.
    line_number : int, the line number to access.
    mask_val : scalar, a value to use in padding.
    sort_order : str, the sort order of the seismic array input.

    Returns
    -------
    traces : 2-D float array containing the trace data for the specified line.
    mask : 2-D boolean array, True value indicated data that has been added by padding.
    """

    sort_order = sort_order.lower()
    if sort_order not in ("inline", "crossline"):
        raise RuntimeError(
            f"{sort_order} not supported, sort order should be on of inline or crossline."
        )

    min_inline, min_crossline, max_inline, max_crossline = bounds

    if sort_order == "inline":
        min_line = min_inline
        max_line = max_inline
    else:
        min_line = min_crossline
        max_line = max_crossline

    if line_number < min_line or line_number > max_line:
        raise RuntimeError(
            f"{line_number} out of bounds [{min_line}, {max_line}]."
        )

    traces = seismic[:, :, line_number - min_line]
    mask = traces < 1
    min_val, max_val = scalers[line_number - min_line, :]
    traces -= 1
    traces = traces.astype(float) * max_val / (65535 - 1)
    traces += min_val
    traces[mask] = mask_val
    return traces, mask


class rssClient:
    def __init__(self, store, cache_size=512 * (1024 ** 2)):
        """
        rss format data access.

        Parameters
        ----------
        store - Instance of s ZArr storage object,
                see s3fs.S3Map for remote s3 storage, or zarr.DirectoryStore as common
                types of store.
        """

        # don't cache meta-data read once
        self.root = zarr.open(store, mode="r")

        clear_output()
        print("Mounting line access.")

        cache = zarr.LRUStoreCache(store, max_size=cache_size)

        inline_root = zarr.open(cache, mode="r")
        self.inline_root = inline_root["inline"]

        crossline_root = zarr.open(cache, mode="r")
        self.crossline_root = crossline_root["crossline"]

        clear_output()
        print("Configuring meta-data.")

        self.bounds = self.root["bounds"]

        self.ilxl = np.vstack(
            [
                self.root["coords"]["inlines"][:],
                self.root["coords"]["crosslines"][:],
            ]
        ).T

        self.xy = np.vstack(
            [self.root["coords"]["cdpx"][:], self.root["coords"]["cdpy"][:]]
        ).T

        self.kdtree = None

        clear_output()
        print("Connection complete.")

    def query_by_xy(self, xy, k=4):
        """
        Query k inline/crossline coordinates closest to this x/y coordinate.

        Parameters
        ----------
        xy - An array containing the [easting, northing] coordinates.
        k - The number of nearest points to look up.

        Returns
        -------
        dist - array, the euclidean distance from the point x/y to the nearest inline/xline grid coordinate.
        ilxl - list, a list of inline/crossling coordinate nearest to the point x/y.
        """

        if self.kdtree is None:
            print(
                "Assembling a tree to map il/xl to x/y. \n"
                + "This could take a couple of minutes, \n"
                + "But only happens one time."
            )
            self.kdtree = KDTree(data=self.xy)

        dist, index = self.kdtree.query(np.atleast_2d(xy), k=k)
        ilxl = [self.ilxl[i, :] for i in index]
        return dist, ilxl

    def line(self, line_number, sort_order="inline"):
        """
        Read a line from the rss data.

        Parameters
        ----------
        line_number : the line number to read.
        sort_order : one of 'inline' or 'crossline' depending on your preference.

        Returns
        -------
        traces : 2-D float array containing the trace data for the specified line.
        mask : 2-D boolean array, False value indicated data that has been added by padding.
        """

        sort_order = sort_order.lower()
        if sort_order not in ("inline", "crossline"):
            raise RuntimeError(
                f"{sort_order} not supported, sort order should be on of inline or crossline."
            )

        if sort_order == "inline":
            seismic = self.inline_root["seismic"]
            scalers = self.inline_root["scalers"]
        else:
            seismic = self.crossline_root["seismic"]
            scalers = self.crossline_root["scalers"]

        return load_line(
            seismic, scalers, self.bounds, line_number, sort_order=sort_order
        )

    def trace(self, inline, crossline):
        """
        Read a line from the rss data.

        Parameters
        ----------
        inline : int, inline coordinate.
        crossline : int, crossline coordinate.

        Returns
        -------
        trace : array, the trace at the coordinates.
        is_live : boolean, is this a live trace
        """

        seismic = self.inline_root["seismic"]
        scalers = self.inline_root["scalers"]

        return load_trace(seismic, scalers, self.bounds, inline, crossline)


class rssFromS3(rssClient):
    def __init__(
        self, filename, client_kwargs=None, cache_size=512 * (1024 ** 2)
    ):
        """
        An object for accessing rss data from s3 blob storage.

        Parameters
        ----------
        filename : path to rss data object on s3.
        client_kwargs : dict containing aws_access_key_id and aws_secret_access_key or None.
        If this variable is none, anonymous access is assumed.
        cache_size : max size of the LRU cache.
        """
        print("Establishing Connection, may take a minute ......")

        anon = client_kwargs is None

        s3 = s3fs.S3FileSystem(anon=anon, client_kwargs=client_kwargs)

        clear_output()
        print("Connected to S3.")

        store = s3fs.S3Map(root=filename, s3=s3, check=False)

        super().__init__(store, cache_size=cache_size)


class rssFromFile(rssClient):
    def __init__(self, filename, cache_size=512 * (1024 ** 2)):
        """
        An object for accessing rss data from s3 blob storage.

        Parameters
        ----------
        filename : path to rss data object on disk.
        """

        store = zarr.DirectoryStore(f"{filename}")
        root = zarr.open(store, mode="r")
        super().__init__(store, cache_size=cache_size)
